import numpy as np
from collections import deque
from umap import UMAP
from sklearn.cluster import HDBSCAN
from sklearn.metrics.pairwise import pairwise_distances
from ..base import GeneratorMixin
from ._cluster_base import Cluster
from ..topic import TopicGenerator
    

class HDBSCANTopicGenerator(GeneratorMixin):

    def __init__(self, reduced_dim, n_neighbors, min_cluster_percent, topk=5, 
                 model_name="gpt-35-turbo", temperature=0.5, verbose=True):
        self.reduced_dim = reduced_dim
        self.n_neighbors = n_neighbors
        self.min_cluster_percent = min_cluster_percent
        self.topk = topk
        self.verbose = verbose
        self.nodes = 1
        self.topic_generator = TopicGenerator(model_name, temperature)

    def _dfs_build_clusters(self, root, n_neighbors, tier):
        if self.verbose:
            print("\t" * tier, "tier:", tier, "#data:", len(root.data), "n_neighbors:", n_neighbors)

        reducer = UMAP(n_neighbors=5, n_components=self.reduced_dim, metric='cosine')
        reduced_embeddings = reducer.fit_transform(np.array(root.data['embeddings'].tolist(), dtype=np.float32))
        
        cluster = HDBSCAN(min_cluster_size=self.min_cluster_size, min_samples=self.n_neighbors, n_jobs=-1)
        cluster.fit(reduced_embeddings)
        root.data['label'] = cluster.labels_
        
        for label in sorted(set(cluster.labels_)):
            cluster_data = root.data[root.data['label'] == label].reset_index(drop=True)
            if self.verbose:
                print("\t" * tier, "cluster label:", label, "#data:", len(cluster_data))

            if len(cluster_data) <= self.min_cluster_size:
                root.children.append(Cluster(data=cluster_data))
            else:
                next_neighbors = min(int(n_neighbors // 2), len(cluster_data))
                if next_neighbors < 2:
                    root.children.append(Cluster(data=cluster_data))
                else:
                    sub_root = self._dfs_build_clusters(Cluster(data=cluster_data), next_neighbors, tier+1)
                    root.children.append(sub_root.children[0] if len(sub_root.children) == 1 else sub_root)
        return root.children[0] if len(root.children) == 1 else root

    def build_cluster_tree(self, data):
        self.min_cluster_size = int(self.min_cluster_percent * len(data))
        root = Cluster(data=data)
        root = self._dfs_build_clusters(root, self.n_neighbors, tier=0)
        return root
        
    def _tree_traversal(self, root):
        if root.size is None:
            root.size = len(root.data)
        if root.percent is None:
            root.percent = 1.

        if root.centroid is None:
            root.centroid = np.mean(np.array(root.data['embeddings'].tolist(), dtype=np.float32), axis=0)
        dists = pairwise_distances(np.array([root.centroid], dtype=np.float32), np.array(root.data['embeddings'].tolist(), dtype=np.float32))[0]
        root.within_cluster_dist = np.mean(dists)
        root.closest = np.argsort(dists)[:self.topk]
        response = self.topic_generator.transform(root)
        root.topic = eval(response)['topic']
        root.description = eval(response)['description']

        for child in root.children:
            child.size = len(child.data)
            child.percent = round(child.size / root.size, 3)
            self._tree_traversal(child)

    def fill_cluster_tree(self, root):
        self._tree_traversal(root)

    def plot_cluster_tree(self, root, last=True, header=''):
        elbow = "└────"
        pipe = "│  "
        tee = "├────"
        blank = "   "
        print(f"{header}{elbow if last else tee} {root.topic} - {root.description} ({root.size} | {root.percent*100:.1f}%)")
        
        child_size = len(root.children)
        self.nodes += child_size
        if child_size > 0:
            for i, c in enumerate(root.children):
                self.plot_cluster_tree(c, header=header + (blank if last else pipe), last=i == child_size - 1)

    def predict(self, data):
        root = self.build_cluster_tree(data)
        self.fill_cluster_tree(root)
        if self.verbose:
            self.plot_cluster_tree(root)
        return root
    
    # def save_to_csv(self, root, save_file="topics.csv"):
    #     results = root.data.copy(deep=True)
        
    #     queue = deque([])
    #     queue.append(root)
    #     level = -1
    #     while queue:
    #         level += 1
    #         level_size = len(queue)
    #         for _ in range(level_size):
    #             node = queue.popleft()

    #             for item in node.data:
    #                 topic_name = "Topic-" + str(level)
    #                 desc_name = "Description-" + str(level)
    #                 condition = results['unique_request_id'] == item['unique_request_id']
    #                 results.loc[condition, topic_name] = item.topic
    #                 results.loc[condition, desc_name] = item.description

    #             for node in node.children:
    #                 queue.append(node)
    #     results = results.drop(columns=['Topic-0', 'Description-0'])
    #     results.to_csv(save_file, index=False)


import re
from ..base import GeneratorMixin
from ..walmart_llm import ChatModel


class Classifier(GeneratorMixin):

    def __init__(self, model_name="gpt-35-turbo", temperature=0., batch_size=300):
        self.model_name = model_name
        self.model = ChatModel(model_name=model_name, temperature=temperature, batch_size=batch_size)

        self.template1 = """You are a ChatGPT user. Your task is to write a prompt asking a question about a topic. You are given some requirements, some examples and one topic below.
    
Here are some requirements you must follow:
1. Your output MUST be less than 100 words.
2. Your output MUST be related about the provided topic.
3. Only output the Topic.
    
Here are some good examples for your reference:
Topic: Acknowledging and forwarding applicant's email
Prompt: I want to thank the applicant for the interest.Would like to express to him that his email will be forwarded to the store manager at 2201.
    
Topic: Listing areas of development
Prompt: now i need to list areas of development
    
Topic: Steps to become a community partner in a health program
Prompt: Use this information to create an agreement form.2.Complete the Community Partnership Questionnaire: Fill out and submit the Community Partnership Questionnaire.This questionnaire will help gather important information about your organization and its capabilities.3.Be a Certified 501(c)(3) Organization: Ensure that your organization is a certified 501(c)(3) nonprofit organization.This certification is necessary to participate in the program as a community partner.4.Register your organization on the FH platform to receive live referrals: Register your organization on the FH platform to receive live referrals.This platform will connect you with individuals (CareSource Members) who have health-related social needs and require assistance.5.Connect individuals to resources to address their health-related social needs: As a community partner, you should be prepared to connect individuals to resources that can address their health-related social needs.
    
Topic: {topic}
Prompt:"""

        self.template2 = """You are a good editor. The following will give you some requirements, a topic list, some examples and a prompt. Your task is to judge which category this prompt belongs to according to the following requirements.

Here are some requirements you must follow:
1. Only output the number of the topic to which the prompt belongs.

Here is the topic list for your reference:
{topics}

Here are some good examples for your reference:
{examples}

Prompt: {prompt}
Topic:"""

    def predict(self, data, topics):
        topic_str = ""
        topic_list = [topic+" - "+desc for topic, desc in topics.items()]
        for i, key in enumerate(topic_list, 1):
            topic_str += (f"[{i}] {key}\n")
        
        prompts1 = [self.template1.format(topic=topic) for topic in topics]
        examples = self.model.batch_completion(prompts1)
        

        topic_examples_str = ""
        for topic, example in zip(topic_list, examples):
            topic_examples_str += f"Prompt: {example[0]}\nTopic: [{topic_list.index(topic)+1}]\n\n"

        prompts = [self.template2.format(prompt=text, topics=topic_str, examples=topic_examples_str) for text in data['input'].tolist()]
        classified_texts = self.model.batch_completion(prompts)

        topic_sz = len(topics)
        data['label'] = [extract_number(text, topic_sz) for text in classified_texts]

        return data

def extract_number(sentence, size):
    pattern = r'\d+'
    match = re.search(pattern, sentence)
    if match:
        num = int(match.group())
        if num in list(range(1, size+1)):
            return int(match.group())
    return 0


from ._cluster_base import Cluster
from ._topic import TopicGenerator
from ._hdbscan_approach import HDBSCANTopicGenerator
from ._kmeans_appproach import KmeansTopicGenerator


__all__ = [
    "Cluster",
    "TopicGenerator",
    "HDBSCANTopicGenerator",
    "KmeansTopicGenerator",
]


import re
import sys
import math
import asyncio
import pandas as pd
from tqdm import tqdm
from azure.ai.textanalytics.aio import TextAnalyticsClient
from azure.ai.textanalytics import PiiEntityCategory
from azure.core.credentials import AzureKeyCredential
from .. import config
from ..base import TransformerMixin


class MinMaxLengthFilter(TransformerMixin):

    def __init__(self, words_range=(0, -1)):
        self.words_range = words_range
        
    def transform(self, data):
        min_words = self.words_range[0]
        max_words = self.words_range[1] if self.words_range[1] != -1 else sys.maxsize

        if isinstance(data, list):
            rt_data = []
            for idx in range(len(data)):
                words = str(data[idx]).strip().split()
                if min_words <= len(words) <= max_words:
                    rt_data.append(data[idx])
                return rt_data
        elif isinstance(data, pd.DataFrame):
            if 'input' not in data.columns:
                raise ValueError(f"Must include `input` column in data.")
            def select_row(row):
                length = len(str(row['input']).strip().split())
                return (length >= min_words) and (length <= max_words)
            rt_data = data[data.apply(select_row, axis=1)]
            rt_data = rt_data.reset_index(drop=True)
            return rt_data
        else:
            raise ValueError("data should be in the type of `list` or `pd.DataFrame`")


class NameFilter(TransformerMixin):

    batch_size = 5

    def __init__(self):
        if config.azure_key is None:
            raise ValueError("Please set up `config.azure_key`!")
        if config.azure_endpoint is None:
            raise ValueError("Please set up `config.azure_endpoint`!")

    async def _name_filter(self, texts):
        client = TextAnalyticsClient(
            endpoint=config.azure_endpoint,
            credential=AzureKeyCredential(config.azure_key),
        )
        async with client:
            response = await client.recognize_pii_entities(documents=texts, language="en", categories_filter = [PiiEntityCategory.PERSON], verify_ssl=False)
            filtered_texts = [re.sub(r'[*]+', 'PersonName', doc.redacted_text) if not doc.is_error else "" for doc in response]
        await client.close()
        return filtered_texts

    def transform(self, data):
        if 'input' not in data.columns:
            raise ValueError(f"Must include `input` column in data.")
        texts = data['input'].tolist()

        rt_texts = []
        num_batch = math.ceil(len(texts) / self.batch_size)
        for idx in tqdm(range(0, len(texts), self.batch_size), total=num_batch):
            rt_texts.extend(asyncio.run(self._name_filter(texts[idx: idx+self.batch_size])))
    
        data['input'] = rt_texts
        return data[data['input'] != ""]


from dataclasses import dataclass


@dataclass
class Config:
    
    # LLM configurations
    consumer_id: str = None
    private_key_path: str = None
    mso_llm_env: str = None

    # Azure PII filter
    azure_key: str = None
    azure_endpoint: str = None


import numpy as np
from sklearn.cluster import KMeans
from sklearn.metrics.pairwise import pairwise_distances
from ..base import GeneratorMixin
from ..topic import Cluster, TopicGenerator
from ..walmart_llm import AdaEmbedModel
from ..embedding import BGEEmbedModel


class KmeansTopicGenerator(GeneratorMixin):

    def __init__(self, n_clusters_list = [], topk=5, batch_size=100, model_name="gpt-35-turbo", 
                 temperature=0.5, embed_model="bge", device="mps", embed_batch_size=300):
        self.n_clusters_list = n_clusters_list
        self.topk = topk
        self.batch_size = batch_size,
        self.topic_generator = TopicGenerator(model_name, temperature, batch_size=batch_size)
        if embed_model == "ada":
            self.embed_model = AdaEmbedModel(batch_size=embed_batch_size)
        elif embed_model == "bge":
            self.embed_model = BGEEmbedModel(batch_size=embed_batch_size, device=device)
        
    def generate_micro_clusters(self, data): 
        kmeans = KMeans(n_clusters=self.n_clusters_list[0], random_state=42, n_init="auto")
        kmeans.fit(np.array(data['embeddings'].tolist(), dtype=np.float32))
        data['label'] = kmeans.labels_
        # generate micro clusters
        clusters = []
        for label in sorted(set(kmeans.labels_)):
            cluster_data = data[data['label'] == label].reset_index(drop=True)
            
            cluster = Cluster(data=cluster_data)
            cluster.centroid = kmeans.cluster_centers_[label]
            cluster.size = len(cluster.data)
            
            dists = pairwise_distances(np.array([cluster.centroid], dtype=np.float32), np.array(cluster.data['embeddings'].tolist(), dtype=np.float32))[0]
            cluster.within_cluster_dist = np.mean(dists)
            cluster.closest = np.argsort(dists)[:self.topk]

            clusters.append(cluster)

        # generate topic and description
        responses = self.topic_generator._generate_batch_topics_and_descriptions_for_clusters(clusters)
        for cluster, response in zip(clusters, responses):
            cluster.topic = eval(response)['topic']
            cluster.description = eval(response)['description']
        return clusters
    
    def generate_higher_clusters(self, clusters, level):
        prompts = []
        for cluster in clusters:
            prompts.append(f"Topic: {cluster.topic}\nDescription: {cluster.description}")
        embeddings = self.embed_model.embed_documents(prompts)
        
        kmeans = KMeans(n_clusters=self.n_clusters_list[level], random_state=42, n_init="auto")
        kmeans.fit(np.array(embeddings, dtype=np.float32))

        higher_clusters = [Cluster() for _ in range(self.n_clusters_list[level])]
        for idx, label in enumerate(kmeans.labels_):
            higher_clusters[label].children.append(clusters[idx])
            
        for cluster in higher_clusters:
            cluster.size = sum([child.size for child in cluster.children])
            for child in cluster.children:
                child.percent = round(child.size / cluster.size, 3)
            
            cluster.keywords = []
            for child in cluster.children:
                cluster.keywords.extend(child.keywords)

        responses = self.topic_generator._generate_batch_topics_and_descriptions_for_higher_clusters(higher_clusters)
        for cluster, response in zip(higher_clusters, responses):
            cluster.topic = eval(response)['topic']
            cluster.description = eval(response)['description']
        return higher_clusters
    
    def predict(self, texts, embeddings, **params):
        clusters = []
        micro_clusters = self.generate_micro_clusters(texts, embeddings)
        clusters.append(micro_clusters)
        for idx in enumerate(self.n_clusters_list[1:], start=1):
            higher_clusters = self.generate_higher_clusters(clusters[-1], level=idx)
            clusters.append(higher_clusters)
        return clusters


import sys
sys.path.append("..")

#-------- Add Config ------------
from topicgpt import config

config.azure_key = "6ea086b8344b41338f484502e3f77cd9"
config.azure_endpoint = "https://my-assistant-text-analyzer-dev-66b3fa72-cognitive.cognitiveservices.azure.com/"

config.consumer_id = "719f6b98-cfb3-4048-9bd0-095b57b3257a"
config.private_key_path = "private_key.pem"
config.mso_llm_env = "stage"

# --------- Load data ------------------
import pandas as pd
data = pd.read_csv("20230725-20240315.csv")
data = data#.sample(n=10000)
data = data.dropna(subset=['userInput'])
texts = data['userInput'].tolist()

# --------- Preprocessing ---------------
from topicgpt.preprocessing import MinMaxLengthFilter

filter = MinMaxLengthFilter((1, 1000))
texts = filter.transform(texts)
print(len(texts))

# -------- Embedding data -----------
from topicgpt.embedding import BGEEmbedModel

llm = BGEEmbedModel(batch_size=500, device="mps")
embeddings = llm.embed_documents(texts)

# -------- Extract Keywords ----------
from topicgpt.feature_extraction import LlmKeywordsExtractor

llm = LlmKeywordsExtractor()
keywords = llm.transform(texts)
breakpoint()

# --------- HDBSCAN -----------------
from topicgpt.topic import HDBSCANTopicGenerator

generator = HDBSCANTopicGenerator(reduced_dim=10, n_neighbors=10, min_cluster_size=50)

# one stop
root = generator.predict(texts, embeddings)

# step by step
# root = generator.build_cluster_tree(texts, embeddings)
# generator.fill_cluster_tree(root)
# generator.plot_cluster_tree(root)


import pandas as pd
from sklearn.feature_extraction.text import TfidfVectorizer
from ..base import TransformerMixin
from ..walmart_llm import ChatModel


class TfidifKeywordsExtractor(TransformerMixin):

    def __init__(self, ngram_range=(1, 2), topk=10):
        self.ngram_range = ngram_range
        self.topk = topk

    def transform(self, clusters):
        # build corpus
        n_classes = len(clusters)
        corpus = [""] * n_classes
        for idx, cluster in enumerate(clusters):
            corpus[idx] += (" ".join(cluster.data['input'].tolist()))

        vectorizer = TfidfVectorizer(ngram_range=self.ngram_range, stop_words='english')
        tfidf = vectorizer.fit_transform(corpus)

        # extract keywords
        feature_names = vectorizer.get_feature_names_out()
        for idx in range(n_classes):
            tfidf_values = tfidf[idx].toarray().flatten()
            sorted_indices = tfidf_values.argsort()[::-1]
            top_keywords_indices = sorted_indices[:self.topk]
            top_keywords = [feature_names[idx] for idx in top_keywords_indices]
            clusters[idx].keywords = ", ".join(top_keywords)
        return clusters


class LlmKeywordsExtractor(TransformerMixin):

    def __init__(self, topk=10, model_name="gpt-35-turbo", temperature=0., batch_size=300):
        self.topk = topk
        self.template = """You will be given a text. This will be enclosed in triple backticks.

Please give me the keywords that are present in this text and separate them with commas.
Make sure you to only return the keywords and say nothing else. For example, don't say: 
"Here are the keywords present in the text". And give at most {topk} keywords.

Text: ```The website mentions that it only takes a couple of days to deliver but I still have not received mine.```
Keywords: deliver, a couple of days, received, website

Text: ```Machine learning (ML) is a field of study in artificial intelligence concerned with the development and study of statistical algorithms that can learn from data and generalize to unseen data, and thus perform tasks without explicit instructions```
Kywords: machine learning, artificial intelligence, statistical algorithms, data, tasks, instructions

Text: ```{text}```
Keywords:
"""
        self.model = ChatModel(model_name=model_name, temperature=temperature, batch_size=batch_size)
    
    def transform(self, data, **params):
        
        def remove_extra_keywords(keywords):
            return [",".join(str(text).strip().split(",")[:self.topk]) for text in keywords]

        if isinstance(data, pd.DataFrame):
            if 'input' not in data.columns:
                raise ValueError(f"Must include `input` column in data.")
            prompts = [self.template.format(topk=self.topk, text=text) for text in data['input'].tolist()]
            keywords = self.model.batch_completion(prompts)
            data['keywords'] = remove_extra_keywords(keywords)
            rt_data = data[data['keywords'] != ''].reset_index(drop=True)
            return rt_data
        elif isinstance(data, list):
            prompts = [self.template.format(topk=self.topk, text=text) for text in data]
            keywords = self.model.batch_completion(prompts)
            return remove_extra_keywords(keywords)
        else:
            raise ValueError("data should be in the type of `list` or `pd.DataFrame`")